#!/usr/bin/perl -w
# Copyright (C) 2006, Augusto Ciuffoletti <augusto@di.unipi.it>
use strict;
use Socket;
use Sys::Hostname;
use IO::Handle;
use POSIX 'setsid';
use bytes;
use stoken;
use peerdb;
use tokenAction;

use Getopt::Std;
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );

my $pidfile="/tmp/agent.pid";

sub cleanup {
	print "*** Cleaning up...\n";
	soft_stop();
	unlink $pidfile;
}

my $VERSION='$Revision: 1.9 $';
$VERSION =~ s/\$//g;
print "agent.pl: $VERSION\n";

if ( -f $pidfile ) {
	open PID,$pidfile;
	my $p=<PID>;
	close PID;
	print "*** Agent already running with pid=$p\n";
	exit 1;
}

my $usage=
"Usage: agent.pl [-l <localhost>] [-p <port>] [-L <logfile>] [-hd] [<peer>]
   Options:
   -l localhost address (default `hostname -f`)
   -p port (default 9000)
   -L alternate logfile (default /tmp/log.out)
   -d daemonize
   -c clear pub keys database and locla keys
   -h this help
   -s time slot for token action in secs\n";
my %opts;
getopts('hl:s:p:dc',\%opts);
(defined $opts{l}) || ( $opts{l}=hostname() );
(defined $opts{s}) || ( $opts{s}=10 );
(defined $opts{p}) || ( $opts{p}=9000 );
(defined $opts{L}) || ( $opts{L}='/tmp/agent.log' );
(defined $opts{h}) && die $usage;
(defined $opts{c}) && clear_pkdb();

# Creating pidfile
open(PID,">",$pidfile);
print PID $$;
close PID;
$SIG{__DIE__} = \&cleanup;

# Create a new handle (do not clean)
my $peer_database= peerdb->new;

printf "Agent starting on IP %s\n using well known port %s\n",$opts{l},$opts{p};
if  ( defined $ARGV[0] ) {
    print " with reference peer $ARGV[0]\n";
} else {
    print " in seed mode\n";
} 

if ( defined $opts{d} ) {
#From man perlipc
               chdir '/'               or die "Can't chdir to /: $!";
               open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
               open STDOUT, '>',$opts{L}  or die "Can't write to $opts{L}: $!";
               defined(my $pid = fork) or die "Can't fork: $!";
               exit if $pid;
               setsid                  or die "Can't start a new session: $!";
               open STDERR, '>&',\*STDOUT or die "Can't write to $opts{L}: $!";
           }

# Initial peer, if indicated on command line
my $first_peer=( defined $ARGV[0] )?$ARGV[0]:undef;

my $pid=open(TO_SOURCE,"|-");
TO_SOURCE->autoflush(1);
( defined $pid ) || die "cannot fork: $!";

if ( $pid ) {
	while ( 1 ) {
#DESTINATION (parent process)
		my ($token_id,$peer,$t1);
		($token_id,$peer,$t1)=token_wait($opts{p},200,10,1);
		if ( ! defined $token_id ) {
			$token_id=randid();
			$peer="newtoken";
			$t1=0;	
		};
#	print "Metto $token_id nel pipe\n";
		print TO_SOURCE "$token_id\n$peer\n";
    }
} else {
#SOURCE (child process)
	if ( defined $first_peer ) {
	    # Starts in "Holding token" state
	    my $peer=$first_peer;
	    my $token_id=randid();
	    # indefinetely tries to send the token to initial peer    
	    while ( ! defined token_pass($token_id,$opts{l},$peer,$opts{p},10,1)) { 
			sleep 10;
	    }
	}
    my $stack=[];
    while (  my $token_id = <STDIN> ) {
		chomp $token_id;
		my $peer=<STDIN>;
		chomp $peer;
		print "Token received from $peer\n";
		$stack=unique($stack,$token_id);
		( $token_id eq $stack->[0] ) || next;
		my $t0=[gettimeofday];
		eval {
	        local $SIG{ALRM} = sub { die "Timeout\n" };
		    alarm $opts{s};
		    action($peer);
			system("ping -c 1 -p".$token_id."AC15 $first_peer");
		    alarm 0;
		};
		warn "Token related action aborted: $@" if $@;
		# keep token for whole time slot (should be an option...)
		( $opts{s} > tv_interval($t0) ) && 
		    sleep( int($opts{s} - tv_interval($t0)) );
		$peer=$peer_database->select_peer();
		( defined $peer ) || die "No more peers in database ?!\n";
		print "== Start passing token $token_id to $peer\n";
		while ( ! defined token_pass ( $token_id, $opts{l}, $peer, $opts{p}, 10, 1 ) ) {
				$peer_database->demote($peer);	
				$peer=$peer_database->select_peer();
			sleep 1;
		    print "== Start passing token $token_id to $peer\n";
		}
		$peer_database->promote($peer);
    }
}
