package peerdb;
use strict;
use Crypt::OpenSSL::DSA;
use Socket;
use Sys::Hostname;

sub new {
	my $class = shift;
	my $self = {};
	bless $self,$class;
	$self->{dbdir}="/tmp/peerdb.d";
#	$self->reset;
	system("mkdir -p $self->{dbdir}") == 0
		or die "system mkdir failed: $?";
	return $self;
}

sub search {
	my $self = shift;
	my ( $peer ) = @_;
	opendir(DB, $self->{dbdir}) 
		|| die "can’t opendir $self->{dbdir}: $!";
	my @fn=grep(/^$peer/,readdir(DB));
	closedir DB;
	# should check and unify duplicates
	return $fn[0];
}
	

sub add {
	my $self = shift;
	my ( $peer, $cert ) = @_;
	my $fn=$self->search($peer);
	if ( defined $fn ) {
		open(OLD,">","$self->{dbdir}/$fn");
		print OLD $cert;
		close OLD;
		$self->promote($peer);
	} else { 
		open(NEW,">","$self->{dbdir}/$peer.0.pub") 
			or die "can’t open $self->{dbdir}/$peer.0.pub: $!";
		print NEW $cert;
		close NEW;
	}
	return 1;
}

sub demote {
	my $self = shift;
	my ( $peer ) = @_;
	my $fn=$self->search($peer);
	( defined $fn ) || return undef;
	my ( $address,$score ) = ( $fn =~ /^(\d+\.\d+\.\d+\.\d+)\.(\d)/ );
	(defined $address ) || die "Malformed database\n";
	$score++;
	my $newname="$address.$score.pub";
	rename($self->{dbdir}."/".$fn,$self->{dbdir}."/".$newname)
		|| die "can’t rename $self->{dbdir}/$fn: $!";
	return 1;
}

sub promote {
	my $self = shift;
	my ( $peer ) = @_;
	my $fn=$self->search($peer);
	( defined $fn ) || return undef;
	# if more than one call unify (internal)
	my ( $address,$score ) = ( $fn =~ /^(\d+\.\d+\.\d+\.\d+)\.(\d)/ );
	(defined $address ) || die "Malformed database\n";
	my $newname="$address.0.pub";
	rename($self->{dbdir}."/".$fn,$self->{dbdir}."/".$newname)
		|| die "can’t rename $self->{dbdir}/$fn: $!";
	return 1;
}


sub get_certificate {
	my $self = shift;
	my ( $peer ) = @_;
	my $fn=$self->search($peer);
	( defined $fn ) || return undef;
    my $cert=Crypt::OpenSSL::DSA->read_pub_key("$self->{dbdir}/$fn");
#	open(FH,"$self->{dbdir}/$fn")
#		or die "Can't open $self->{dbdir}/$fn: $!\n";
#	my $cert=<FH>;
	close FH;
	return $cert;
}

sub get_peers {
	my $self = shift;
	opendir(DB, $self->{dbdir}) 
		|| die "can’t opendir $self->{dbdir}: $!";
	my @fn = grep(/.pub$/,readdir(DB));
	for (my $i=0; $i<=$#fn; $i++) {
		my ( $ip, $score ) = ( $fn[$i] =~ /^(\d+\.\d+\.\d+\.\d+)\.(\d)/ );
#		print "gp $i, $fn[$i], $ip, $score\n";
		$fn[$i]= [ $ip, $score ];
	}
	closedir(DB);
	return \@fn;
}

sub reset {
	my $self = shift;
	system("rm -Rf $self->{dbdir}") == 0
		or die "Database reset failed";
	system("mkdir -p $self->{dbdir}") == 0
		or die "system mkdir failed: $?";
	open(FH,"peer")
		or die "Can't open initial peer file: $!\n";;
	while ( my $p=<FH> ) {
		chomp $p;
		open(PFH,">","$self->{dbdir}/$p.0.pub")
 			or die "Can't create $self->{dbdir}/$p.0.pub: $?";;
		close PFH;
	}
}

sub select_peer {
	my $self = shift;
	sub byscore {
		my ( $sa ) = $a->[1];
		my ( $sb ) = $b->[1];
		return $sb <=> $sa;
	}
	sub triangular {
# use a triangular distribution to select peers,
# with a long constant distribution and a steep
# slope for unreliable peers
		my $c=0.9;
		my $u=rand(1);
		if ( $u < $c ) {
			return sqrt($c*$u);
		} else {
			return 1-sqrt((1-$c)*(1-$u));
		}
	}
	my $listref=$self->get_peers;	
	my @sorted = sort byscore @$listref;
	my ( $peer ) = ( $sorted[ int(scalar(@sorted)*triangular) ] )->[0];
	return $peer;
}

sub sync {
	my $self = shift;
	my ( $listref ) = @_;
	my $iaddr=gethostbyname(hostname());
	my $localhost=inet_ntoa($iaddr);
	my @remote=sort { $a cmp $b } @$listref;
	opendir(DB, $self->{dbdir}) 
		|| die "can’t opendir $self->{dbdir}: $!";
	my @fn = grep(/.pub$/,readdir(DB));
	closedir(DB);
	my @local=sort { $a cmp $b } @fn;
	my ( $xl, $xr ) = ( 0, 0);
	while ( $xr le $#remote ) {
		if ( $remote[$xr] eq $localhost ) {
			$xr++;
			next;
		}
		if ( ! defined $local[$xl] ) { 
			print "EOL Adding new peer $remote[$xr]"; print "\n";
			open(NEW,">","$self->{dbdir}/$remote[$xr].0.pub") 
				or die "can’t open $self->{dbdir}/$remote[$xr].0.pub: $!";
			close NEW;
			$xr++;
			next;
		}
        my ( $n ) = ( $local[$xl] =~ /^(\d+\.\d+\.\d+\.\d+)/ );
		print "check $n  $remote[$xr] ";
		if ( $n lt $remote[$xr] ) { $xl++; print "lt\n"; next }
		if ( $n eq $remote[$xr] ) {	 $xr++; $xl++; print "eq \n"; next }
		print "gt\n";
		print "Adding new peer $remote[$xr]"; print "\n";
		open(NEW,">","$self->{dbdir}/$remote[$xr].0.pub") 
			or die "can’t open $self->{dbdir}/$remote[$xr].0.pub: $!";
		close NEW;
		$xr++;
	}
}
	

return 1;
