#!/usr/bin/perl -w
use strict;
# Copyright (C) 2006, Augusto Ciuffoletti <augusto@di.unipi.it>
my $VERSION='$Revision: 1.5 $';
$VERSION =~ s/\$//g;
$|=1;

use Socket;
use Carp;
use Sys::Hostname;
use POSIX 'setsid';
use peerdb;

sub logmsg { print "$0 $$: @_ at ", scalar localtime, "\n" }

my $logfile="/tmp/server.log";
#my $certdir="/tmp";
my $pidfile="/tmp/server.pid";

my $iaddr=gethostbyname(hostname());
my $localhost=inet_ntoa($iaddr);
my $port = 9001;
my $proto = getprotobyname('tcp');
my $peer_database=peerdb->new;

sub cleanup {
	logmsg "Cleaning up... ";
	unlink $pidfile;
}

if ( -f $pidfile ) {
	open PID,$pidfile;
	my $p=<PID>;
	close PID;
	print "*** Server already running with pid=$p\n";
	exit 1;
} 

# It is BETTER to chdir, but next we need to configure the location of
# config files (seed etc)
#chdir '/'               
#    or die "Can't chdir to /: $!";
open STDIN, '/dev/null' 
    or die "Can't read /dev/null: $!";
open STDOUT, '>',$logfile or die "Can't write to $logfile: $!";
defined(my $pid = fork) 
    or die "Can't fork: $!";
exit if $pid;
setsid                  or die "Can't start a new session: $!";
open STDERR, '>&',\*STDOUT or die "Can't write to $logfile: $!";

socket(Server, PF_INET, SOCK_STREAM, $proto)        
    || die "socket: $!";
setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))   
    || die "setsockopt: $!";
bind(Server, sockaddr_in($port, inet_aton($localhost)))        
     || die "bind: $!";
listen(Server,SOMAXCONN)                            
     || die "listen: $!";

logmsg "server started on port $port as $localhost";
my $paddr;

# Reset database
$peer_database->reset;

$SIG{CHLD} = \&REAPER;
# Creating pidfile
open(PID,">",$pidfile);
print PID $$;
close PID;
$SIG{TERM} = \&cleanup;
$SIG{INT} = \&cleanup;

for ( ; $paddr = accept(Client,Server); close Client) {
    my($port,$iaddr) = sockaddr_in($paddr);
#    my $name = gethostbyaddr($iaddr,AF_INET);
    
    logmsg "connection from ", inet_ntoa($iaddr), " at port $port";
    
	my @peerdb=();
    while ( defined(my $p = <Client>) ) {
		chomp $p;
		push @peerdb,$p
	}
	$peer_database->sync(\@peerdb);
}
